local _, ns = ...
local OnCast = ns.addon
local Event = OnCast:NewModule("Event", "AceEvent-3.0")

-- Lua APIs
local mceil = math.ceil
local mfmod = math.fmod
local sformat = string.format

-- WoW APIs
local WOW = OnCast.WOWAPI or {}
WOW.isBCC = (WOW_PROJECT_ID == WOW_PROJECT_BURNING_CRUSADE_CLASSIC)
WOW.GetTime = GetTime
WOW.GetSpellInfo = GetSpellInfo -- https://wow.gamepedia.com/API_GetSpellInfo
WOW.UnitCastingInfo = UnitCastingInfo -- https://wow.gamepedia.com/API_UnitCastingInfo
if WOW.isBCC then
    WOW.PlayerCastTiming = function()
        local _,_,_,startTime,endTime,_,guid,id = WOW.UnitCastingInfo("player")
        return startTime,endTime,guid,id
    end
else
    WOW.PlayerCastTiming = function()
        local _,_,_,startTime,endTime,_,guid,_,id = WOW.UnitCastingInfo("player")
        return startTime,endTime,guid,id
    end
end

WOW.UnitChannelInfo = UnitChannelInfo -- https://wow.gamepedia.com/API_UnitChannelInfo
if WOW.isBCC then
    WOW.PlayerChannelTiming = function()
        local _,_,_,startTime,endTime,_,id = WOW.UnitChannelInfo("player")
        return startTime,endTime,id
    end
else
    WOW.PlayerChannelTiming = function()
        local _,_,_,startTime,endTime,_,_,id = WOW.UnitChannelInfo("player")
        return startTime,endTime,id
    end
end
WOW.GetUnitName = GetUnitName -- https://wow.gamepedia.com/API_GetUnitName

-- Modules
local Util = nil
local ExtEvent = nil

-- Events
local EVENT_TYPE = {
    START = "START",
    STOP = "STOP",
    DELAYED = "DELAYED",
    FAILED = "FAILED",
    INTERRUPTED = "INTERRUPTED",
    SUCCEEDED = "SUCCEEDED",
    -- Special Events
    FAILED_COOLDOWN = "FAILED_COOLDOWN",
    FAILED_LOSSOFCONTROL = "FAILED_LOSSOFCONTROL"
}

-- Event Aliases
local SUPPORTED_EVENTS = {
    ["START"] = EVENT_TYPE.START,
    ["STARTED"] = EVENT_TYPE.START,
    ["SR"] = EVENT_TYPE.START,
    ["STOP"] = EVENT_TYPE.STOP,
    ["STOPPED"] = EVENT_TYPE.STOP,
    ["SP"] = EVENT_TYPE.START,
    ["DELAYED"] = EVENT_TYPE.DELAYED,
    ["DELAY"] = EVENT_TYPE.DELAYED,
    ["DL"] = EVENT_TYPE.DELAYED,
    ["FAILED"] = EVENT_TYPE.FAILED,
    ["FAILURE"] = EVENT_TYPE.FAILED,
    ["FAIL"] = EVENT_TYPE.FAILED,
    ["FL"] = EVENT_TYPE.FAILED,
    ["INTERRUPTED"] = EVENT_TYPE.INTERRUPTED,
    ["INTERRUPT"] = EVENT_TYPE.INTERRUPTED,
    ["IN"] = EVENT_TYPE.INTERRUPTED,
    ["SUCCEEDED"] = EVENT_TYPE.SUCCEEDED,
    ["SUCCESS"] = EVENT_TYPE.SUCCEEDED,
    ["SU"] = EVENT_TYPE.SUCCEEDED,
}
OnCast.SUPPORTED_EVENTS = SUPPORTED_EVENTS

-- Holds values for the currently casting or channeling spell
local _CastData = nil
-- Holds a map of recent casts SENT data by GUID
local _CastsSent = {}
local _GuidBySpellId = {}
-- Index cast sent data in order
local _sentIndex = {}
local _sentIdxMin, _sentIdxMax = 1, 0

local function wipeData()
    _CastData = nil
    wipe(_CastsSent)
    wipe(_GuidBySpellId)
    wipe(_sentIndex)
    _sentIdxMin, _sentIdxMax = 1, 0
end

-- Module functions
function Event:OnInitialize()
    Util = OnCast:GetModule("Util")
    ExtEvent = OnCast:GetModule("ExtEvent")
end

function Event:OnEnable()
    wipeData()

    -- Subscribe to events
    self:RegisterEvent("CURRENT_SPELL_CAST_CHANGED", Event.OnCastCancelled)
    self:RegisterEvent("UNIT_SPELLCAST_SENT", Event.OnCastSent);
    self:RegisterEvent("UNIT_SPELLCAST_FAILED", Event.OnCastFailed);
    self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", Event.OnCastSucceeded);
    self:RegisterEvent("UNIT_SPELLCAST_START", Event.OnCastStarted);
    self:RegisterEvent("UNIT_SPELLCAST_STOP", Event.OnCastEvent);
    self:RegisterEvent("UNIT_SPELLCAST_DELAYED", Event.OnCastEvent);
    self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", Event.OnCastEvent);
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", Event.OnChannelStarted);
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", Event.OnChannelEvent);
    self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE", Event.OnChannelEvent);
    
end

function Event:OnDisable()
    -- Unsubscribe all events
    self:UnregisterAllEvents()
    
    wipeData()
end

-- Monitor for cancelled cast targetting; e.g. AoE Ground target spells
function Event.OnCastCancelled(eventName, isCancelled)
    -- Only interrested in cancelled events
    if not isCancelled then return end

    if OnCast.debug then
        OnCast.printDebugMessage(sformat("%s - Spell Cast Cancelled", eventName))
    end

    OnCast.Announce = nil -- Reset so we don't get shadowed messages
end

-- Event Order for Spells --
-- NORMAL: SENT > START > [DELAYED] > [SUCCEEDED|INTERRUPTED|FAILED] > STOP
-- INSTANT: SENT > SUCCEEDED
-- CHANNEL: SENT > START > SUCCEEDED > [SUCCEEDED] > STOP
-- ANY: [SENT], FAILED

-- The type of cast; certain types use different functions
local CAST_TYPE = {
    NORMAL = 1,
    INSTANT = 2,
    CHANNEL = 3
}

-- A list of spells to ignore
local IGNORE_SPELL = {
    [75] = true, -- Auto Shot
    [2480] = true, -- Shoot Bow
    [2764] = true, -- Throw
    [5019] = true, -- Shoot (Wand)
    [7918] = true, -- Shoot Gun
    [7919] = true, -- Shoot Crossbow
    [22009] = true, -- Priest 8-set 'Greater Heal' HoT
}
Event.IGNORE_SPELL = IGNORE_SPELL

-- Translate channel effect spell id back to the original spell id
local TRANSLATE_SPELLID = {
    [20577] = 20578 -- Cannibalize
}

-- Cast SENT triggers at the beginning of almost* every cast attempt.
-- This includes attempted casts while already casting!
-- * Some casts will instantly fail without a preceeding SENT event
function Event.OnCastSent(eventName, unitId, targetName, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local spellName, _, _, spellCastTime = WOW.GetSpellInfo(spellId)
        local casterName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        OnCast.printDebugMessage(sformat("%s - Caster:(%s) %s - Target: %s - Spell:(%s) %s @ %s - Guid: %s",
            eventName, unitId, casterName, targetName or "N/A", spellId, spellName, spellCastTime or 0, guid))
    end
    
    local time = WOW.GetTime()

    local announce = OnCast.Announce
    OnCast.Announce = nil -- Reset so we don't get shadowed messages

    -- Check timeout
    if announce and (time - announce.timestamp) > announce.timeout then
        announce = nil
    end

    -- Check for registered spells
    if not announce then
        local spellName = WOW.GetSpellInfo(spellId)
        announce = Util.tableDeepCopy(OnCast.SpellAnnounce[spellId] or OnCast.SpellAnnounce[spellName])
    end

    -- Clean up old data
    for i=_sentIdxMin, _sentIdxMax do
        local sentId = _sentIndex[i]
        local data = _CastsSent[sentId]
        if (time - data.timestamp) > 2 then -- 2 Second is more than long enough between SENT and START / SUCCEEDED / FAILED events
            _CastsSent[sentId] = nil
            _sentIndex[i] = nil
            _sentIdxMin = _sentIdxMin + 1
        else
            break -- Indexed in order, so break early
        end
    end
    if _sentIdxMin > _sentIdxMax then
        _sentIdxMin, _sentIdxMax = 1, 0 -- Rebase
    end

    -- Check for the previously set announcement
    if not announce then
        local sent = _CastsSent[_GuidBySpellId[spellId]]
        announce = sent and sent.announcement
    end

    -- Return if we have not announcement
    if not announce then return end

    -- Create SENT event data
    local sentData = {
        timestamp = time,
        targetName = targetName,
        announcement = announce,
    }

    -- Map sent data
    _CastsSent[castGuid] = sentData
    _GuidBySpellId[spellId] = castGuid
    -- Index the guid
    _sentIdxMax = _sentIdxMax + 1
    _sentIndex[_sentIdxMax] = castGuid
end

function Event.DebugLists()
    OnCast.printInfoMessage("Sent Announcements:")
    for k,v in pairs(_CastsSent) do
        OnCast.printInfoMessage(sformat(" - %s -Age: %0.2fs", strsub(k,23), (WOW.GetTime() - v.timestamp)))
    end
end

-- SUCCEEDED -- 
-- Called when the current cast succeeds
local function currentCastSucceeded(eventName, unitId, castGuid, spellId)
    -- Update event and outcome
    _CastData.event = "SUCCEEDED"
    _CastData.outcome = "Succeeded"
    
    -- Reference to local
    local cast = _CastData

    -- Set extended events
    if cast.announcement.meta and cast.announcement.meta.hasExtendedEvents then
        ExtEvent.MonitorSpell(cast.announcement, cast)
    end

    -- Check for announcement
    local onSucceeded = cast.announcement[EVENT_TYPE.SUCCEEDED]
    if not onSucceeded then return end

    -- Only announce Succeeded once per cast
    cast.announcement[EVENT_TYPE.SUCCEEDED] = nil
    
    OnCast.outputSpellCast(onSucceeded, cast)
end

-- Called when a cast instantly succeeds
local function instantCastSucceeded(eventName, unitId, castGuid, spellId)
    -- Check for sent data
    local sent = _CastsSent[castGuid]
    if not sent then return end

    -- Translate certain spells to the continued effect id
    if TRANSLATE_SPELLID[spellId] then
        _GuidBySpellId[TRANSLATE_SPELLID[spellId]] = castGuid
    end
    
    -- Instant success is a capture event, so clear last guid
    _GuidBySpellId[spellId] = nil

    local hasExtendedEvents = sent.announcement.meta and sent.announcement.meta.hasExtendedEvents

    -- Check for announcement
    local onSucceeded = sent.announcement[EVENT_TYPE.SUCCEEDED]
    if not (onSucceeded or hasExtendedEvents) then return end

    -- Get information about the spell
    local spellName, _, spellIcon = WOW.GetSpellInfo(spellId)

    -- Create cast data
    local cast = {}
    cast.type = CAST_TYPE.INSTANT -- Default to Instant
    cast.event = "SUCCEEDED"
    cast.guid = castGuid
    cast.spell = {id = spellId, name = spellName, icon = spellIcon}
    cast.outcome = "Succeeded"
    cast.targetName = sent.targetName

    -- Set extended events
    if hasExtendedEvents then
        ExtEvent.MonitorSpell(sent.announcement, cast)
    end
    
    if onSucceeded then
        OnCast.outputSpellCast(onSucceeded, cast)
    end
end

-- Cast SUCCEEDED triggers when a spell is successfully cast
-- This can happen instantly or for the current cast
function Event.OnCastSucceeded(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for a current cast
    if _CastData and _CastData.guid == castGuid then
        currentCastSucceeded(eventName, unitId, castGuid, spellId)
    else
        instantCastSucceeded(eventName, unitId, castGuid, spellId)
    end
end

-- FAILED --
-- Called when the current cast fails
local function currentCastFailed(eventName, unitId, castGuid, spellId)
    -- Update event and outcome
    _CastData.event = "FAILED"
    _CastData.outcome = "Failed"
    
    -- Reference to local
    local cast = _CastData

    -- Check for announcement
    local onFailed = cast.announcement[EVENT_TYPE.FAILED]
    if not onFailed then return end

    -- Only announce Failed once per cast
    cast.announcement[EVENT_TYPE.FAILED] = nil

    -- Set failure reason
    cast.reason = ExtEvent.GetLastFailureReason(cast.spell.name, cast.spell.id)
    --cast.reason = failureReason(eventName, spellId)

    OnCast.outputSpellCast(onFailed, cast)
end

local function formatTime(time)
    if time < 10 then
        return sformat("%0.1fs", time)
    end
    
    local itime = mceil(time)

    if itime < 70 then
        return sformat("%ds", itime)
    end
    
    local secs = mfmod(itime,60)
    local mins = (itime - secs) / 60
    return sformat("%dm%ds", mins, secs)
end

-- Called when a cast instantly fails
-- NOT always preceeded by a SENT event
local function instantCastFailed(eventName, unitId, castGuid, spellId)
    -- Check for sent data
    local sent = _CastsSent[castGuid]

    -- Get information about the spell
    local spellName, _, spellIcon = WOW.GetSpellInfo(spellId)

    -- Search for announcement
    local announce = nil
    if sent then
        announce = sent.announcement
        -- Instant fail is a capture event, so clear last guid
        _GuidBySpellId[spellId] = nil
    else
        -- Get announcement from macro
        announce = OnCast.Announce
        OnCast.Announce = nil -- Reset this so we don't get shadowed messages

        -- Check timeout
        if announce and (WOW.GetTime() - announce.timestamp) > announce.timeout then
            announce = nil
        end

        -- Check for registered spells
        if not announce then
            announce = Util.tableDeepCopy(OnCast.SpellAnnounce[spellId] or OnCast.SpellAnnounce[spellName])
        end
    end

    -- Check failed events
    if announce and (announce[EVENT_TYPE.FAILED] or announce[EVENT_TYPE.FAILED_LOSSOFCONTROL] or announce[EVENT_TYPE.FAILED_COOLDOWN]) then
        local reason, delta, type = ExtEvent.GetLastFailureReason(spellName, spellId)
        
        local onFailed = nil
        -- Loss of Control
        if type == "lossofcontrol" then
            onFailed = announce[EVENT_TYPE.FAILED_LOSSOFCONTROL]
            if onFailed and onFailed.vars then
                local vars = onFailed.vars
                if vars.minimum and delta < vars.minimum then
                    onFailed = nil
                elseif vars.maximum and delta > vars.maximum then 
                    onFailed = nil
                end
            end
        -- On Cooldown
        elseif type == "cooldown" then
            onFailed = announce[EVENT_TYPE.FAILED_COOLDOWN]
            if onFailed then
                if onFailed.vars then
                    local vars = onFailed.vars
                    if delta < (vars.minimum or 2) then
                        onFailed = nil
                    elseif vars.maximum and delta > vars.maximum then 
                        onFailed = nil
                    end
                elseif delta < 2 then
                    onFailed = nil
                end
            end
        end
        -- Fallback to failed
        if not onFailed then
            onFailed = announce[EVENT_TYPE.FAILED]
            if delta then
                reason = sformat("%s (%s)", reason, formatTime(delta))
                delta = nil
            end
        end

        -- Only announce Failed once per cast sent
        announce[EVENT_TYPE.FAILED_LOSSOFCONTROL] = nil
        announce[EVENT_TYPE.FAILED_COOLDOWN] = nil
        announce[EVENT_TYPE.FAILED] = nil

        if not onFailed then return end

        -- Create cast data
        local cast = {}
        cast.type = CAST_TYPE.INSTANT -- Default to Instant
        cast.event = "FAILED"
        cast.guid = castGuid
        cast.spell = {id = spellId, name = spellName, icon = spellIcon}
        cast.outcome = "Failed"
        cast.targetName = sent and sent.targetName
        
        -- Set failure vars
        cast.reason = reason
        cast.cooldown = delta and formatTime(delta)
        
        OnCast.outputSpellCast(onFailed, cast)   
    end
end

-- Cast FAILED triggers when a spell fails to cast (out of range, etc.)
-- This can happen instantly or for the current cast
function Event.OnCastFailed(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for a current cast
    if _CastData and _CastData.guid == castGuid then
        currentCastFailed(eventName, unitId, castGuid, spellId)
    else
        instantCastFailed(eventName, unitId, castGuid, spellId)
    end

    OnCast.Announce = nil -- Reset this so we don't get shadowed messages after failed casts
end

-- Cast START triggers for Normal spells when they actually begin to cast
function Event.OnCastStarted(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for sent
    local sent = _CastsSent[castGuid]
    if not sent then return end
    
    -- Started is a capture event, so clear last guid
    _GuidBySpellId[spellId] = nil

    -- Get information about the spell
    local spellName, _, spellIcon = WOW.GetSpellInfo(spellId)

    -- Create Cast Data
    local cast = {}
    cast.type = CAST_TYPE.NORMAL
    cast.event = "START"
    cast.guid = castGuid
    cast.spell = {id = spellId, name = spellName, icon = spellIcon}
    cast.targetName = sent.targetName

    -- Set announcement
    cast.announcement = sent.announcement
    
    -- Set cast timings
    local startTime,endTime,guid = WOW.PlayerCastTiming()
    if startTime and endTime and (cast.guid == guid) then
        cast.spell.startTime = startTime
        cast.spell.endTime = endTime
        cast.spell.castTime = endTime - startTime
    end
    
    -- Set cast data for follow-up events
    _CastData = cast

    -- Check for announcement
    local onStarted = cast.announcement[EVENT_TYPE.START]
    if not onStarted then return end

    -- Only announce Started once per cast
    cast.announcement[EVENT_TYPE.START] = nil

    OnCast.outputSpellCast(onStarted, cast)
end

-- Channel START triggers for Channelled spells when they actually begin to cast
function Event.OnChannelStarted(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for sent. Channelled events don't seem to pass a GUID.
    local guid = castGuid or _GuidBySpellId[spellId]
    local sent = _CastsSent[guid]
    if not sent then return end

    -- Started is a capture event, so clear last guid
    _GuidBySpellId[spellId] = nil

    -- Get information about the spell
    local spellName, _, spellIcon = WOW.GetSpellInfo(spellId)

    -- Create Cast Data
    local cast = {}
    cast.type = CAST_TYPE.CHANNEL
    cast.event = "START"
    cast.guid = guid
    cast.spell = {id = spellId, name = spellName, icon = spellIcon}
    cast.targetName = sent.targetName

    -- Set announcement
    cast.announcement = sent.announcement

    -- Set cast timings
    local startTime,endTime,id = WOW.PlayerChannelTiming()
    if startTime and endTime and (cast.spell.id == id) then
        cast.spell.startTime = startTime
        cast.spell.endTime = endTime
        cast.spell.castTime = endTime - startTime
    end

    -- Set cast data for follow-up events
    _CastData = cast

    -- Check for announcement
    local onStarted = cast.announcement[EVENT_TYPE.START]
    if not onStarted then return end

     -- Only announce Started once per cast
    cast.announcement[EVENT_TYPE.START] = nil

    OnCast.outputSpellCast(onStarted, cast)
end

-- Populate to bind spell events to functions
local BIND_EVENT = {}

-- General Cast Event Handler, checks unitId is player, spell ignore, matches guid, and debug
function Event.OnCastEvent(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for a matching cast GUID
    if not (_CastData and _CastData.guid == castGuid) then 
        return
    end

    BIND_EVENT[eventName](eventName, unitId, castGuid, spellId)
end

-- General Channel Event Handler, checks unitId is player, spell ignore, matches id, and debug
function Event.OnChannelEvent(eventName, unitId, castGuid, spellId)
    -- Only look for the players spells that aren't ignored
    if unitId ~= "player" or IGNORE_SPELL[spellId] then return end

    if OnCast.debug then
        local guid = castGuid or "???"
        local unitName = unitId and WOW.GetUnitName(unitId,true) or "N/A"
        local spellName = spellId and (WOW.GetSpellInfo(spellId)) or "Unknown"
        OnCast.printDebugMessage(sformat("%s - Unit: (%s)%s - Spell: (%s)%s - Guid: %s", 
            eventName, unitId, unitName, spellId, spellName, guid))
    end

    -- Check for a matching spell id... apparently Channelled spells don't pass their GUID
    if not (_CastData and _CastData.spell.id == spellId) then 
        return 
    end

    BIND_EVENT[eventName](eventName, unitId, castGuid, spellId)
end

-- Cast STOP triggers for Normal spells when they finish casting
-- This is always preceeded by a START event.
local function castStopped(eventName, unitId, castGuid, spellId)
    -- Update event
    _CastData.event = "STOP"

    -- Reference to local
    local cast = _CastData

    -- STOP is a final event so clear cast data
    _CastData = nil

    -- Check for announcement
    local onStopped = cast.announcement[EVENT_TYPE.STOP]
    if not onStopped then return end
    
    OnCast.outputSpellCast(onStopped, cast)
end
BIND_EVENT["UNIT_SPELLCAST_STOP"] = castStopped

-- Channel STOP triggers for Channelled spells when they stop casting
-- This is always preceeded by a START event.
local function channelStopped(eventName, unitId, castGuid, spellId)
    -- Update event
    _CastData.event = "STOP"

    -- Calculate final casting time for channelled spells (they don't get interrupted events)
    if _CastData.spell.startTime and _CastData.spell.endTime then
        local endTime = WOW.GetTime() * 1000 -- UnitChannelInfo returns nil, so use WOW.GetTime() as an estimate
        local delay = _CastData.spell.endTime - endTime
        if delay > 50 then -- Don't report differences within rounding error
            _CastData.spell.endTime = endTime
            _CastData.spell.castTime = endTime - _CastData.spell.startTime
            _CastData.spell.delay = (_CastData.spell.delay or 0) + delay
        end
    end

    -- Reference to local
    local cast = _CastData

    -- STOP is a final event so clear cast data
    _CastData = nil

    -- Check for announcement
    local onStopped = cast.announcement[EVENT_TYPE.STOP]
    if not onStopped then return end
    
    OnCast.outputSpellCast(onStopped, cast)
end
BIND_EVENT["UNIT_SPELLCAST_CHANNEL_STOP"] = channelStopped

-- Cast DELAYED triggers when a spell cast is knocked back (e.g. by damage taken)
-- This happens between START and STOP events
local function castDelayed(eventName, unitId, castGuid, spellId)    
    -- Update event
    _CastData.event = "DELAYED"

    -- Calculate the delay for a matching cast
    local startTime,endTime,guid = WOW.PlayerCastTiming()    
    if startTime and endTime and _CastData.spell.startTime and _CastData.spell.endTime and (_CastData.guid == guid) then
        local delay = endTime - _CastData.spell.endTime
        if delay < 50 then return end -- Prevent spam of 'zero' second delays
        _CastData.spell.endTime = endTime
        _CastData.spell.castTime = endTime - _CastData.spell.startTime
        _CastData.spell.delay = (_CastData.spell.delay or 0) + delay
    end

    -- Reference to local
    local cast = _CastData

    -- Check for announcement
    local onDelayed = cast.announcement[EVENT_TYPE.DELAYED]
    if not onDelayed then return end
    
    OnCast.outputSpellCast(onDelayed, cast)
end
BIND_EVENT["UNIT_SPELLCAST_DELAYED"] = castDelayed

-- Channel UPDATED triggers when a channelled spell updates (i.e. gets 'delayed')
-- This happens between START and STOP events
local function channelUpdated(eventName, unitId, castGuid, spellId)
    -- Update event
    _CastData.event = "CLIPPED"

    -- Calculate the delay for a matching cast (based on spellId for channelled spells)
    local startTime,endTime,id = WOW.PlayerChannelTiming()
    if startTime and endTime and _CastData.spell.startTime and _CastData.spell.endTime and (_CastData.spell.id == id) then
        local delay = _CastData.spell.endTime - endTime
        _CastData.spell.endTime = endTime
        _CastData.spell.castTime = endTime - _CastData.spell.startTime
        _CastData.spell.delay = (_CastData.spell.delay or 0) + delay
    end

    -- Reference to local
    local cast = _CastData

    -- Check for announcement
    local onUpdated = cast.announcement[EVENT_TYPE.DELAYED]
    if not onUpdated then return end
    
    OnCast.outputSpellCast(onUpdated, cast)
end
BIND_EVENT["UNIT_SPELLCAST_CHANNEL_UPDATE"] = channelUpdated

-- Cast INTERRUPTED triggers when a spell cast is interrupted, such as from moving
-- This happens between START and STOP events
local function castInterrupted(eventName, unitId, castGuid, spellId)
    -- Update event and outcome
    _CastData.event = "INTERRUPTED"
    _CastData.outcome = "Interrupted"

    -- Calculate final cast timings
    if _CastData.spell.startTime and _CastData.spell.endTime then
        local endTime = WOW.GetTime() * 1000 -- UnitCastingInfo returns nil, so use WOW.GetTime() as an estimate
        local delay = endTime - _CastData.spell.endTime
        _CastData.spell.endTime = endTime
        _CastData.spell.castTime = endTime - _CastData.spell.startTime
        _CastData.spell.delay = (_CastData.spell.delay or 0) + delay
    end

    -- Reference to local
    local cast = _CastData

    -- Check for announcement
    local onInterrupted = cast.announcement[EVENT_TYPE.INTERRUPTED]
    if not onInterrupted then return end
    
     -- Only announce Interrupted once per cast
    cast.announcement[EVENT_TYPE.INTERRUPTED] = nil

    OnCast.outputSpellCast(onInterrupted, cast)
end
BIND_EVENT["UNIT_SPELLCAST_INTERRUPTED"] = castInterrupted
