local _, ns = ...
ns.addon = {}
local OnCast = LibStub("AceAddon-3.0"):NewAddon(ns.addon, "OnCast", "AceConsole-3.0")

-- Version
OnCast.VERSION = {KEY="010705r", MAJOR=1, MINOR=7, PATCH=5}
-- Statics
OnCast.HEADING_COLOR = "9999ff"

-- Chat commands
local COMMAND = "oncast"
local MACRO_COMMAND = "oncast-m"
local MACRO_COMMAND_SHORT = "oc-m"
local SPELL_COMMAND = "oncast-spell"
local DEBUG_COMMAND = "oncast-debug"

-- Lua APIs
local pcall = pcall
local print = print
local tonumber = tonumber
local sformat = string.format
local smatch = string.match 
local tconcat = table.concat
local pairs = pairs
local function pack(...)
    return {...}
end

-- WoW APIs
OnCast.WOWAPI = {}
local WOW = OnCast.WOWAPI
WOW.GetTime = GetTime
WOW.GetSpellInfo = GetSpellInfo  -- https://wow.gamepedia.com/API_GetSpellInfo
WOW.GetRaidTargetIndex = GetRaidTargetIndex -- https://wowpedia.fandom.com/wiki/API_GetRaidTargetIndex
WOW.UnitExists = UnitExists -- https://wowpedia.fandom.com/wiki/API_UnitExists
WOW.GetUnitName = GetUnitName -- https://wowpedia.fandom.com/wiki/API_GetUnitName

-- Modules
local Console = nil
local Util = nil
local MacroConditionals = nil

-- Database default values
local db_defaults = {
    global = {
        version = OnCast.VERSION
    },
    char = {
        spell_announce = {},
    }
}

-- Module functions
function OnCast:OnInitialize()
    --DataBase
    self.db = LibStub("AceDB-3.0"):New("OnCastDB", db_defaults)
    OnCast.SpellAnnounce = self.db.char.spell_announce

    -- Get Modules
    Console = OnCast:GetModule("Console")
    Util = OnCast:GetModule("Util")
    MacroConditionals = OnCast:GetModule("MacroConditionals")

    -- Version Update
    local dbVKey = self.db.global.version.KEY
    if not (dbVKey == OnCast.VERSION.KEY) then
        -- Update INSTANCE_CHAT to BATTLEGROUND
        if dbVKey and (#dbVKey < 7 or dbVKey < "010600a") then
            OnCast.printInfoMessage(sformat("Version Update: %s to %s", dbVKey, OnCast.VERSION.KEY), OnCast.HEADING_COLOR)
            for spn, sp in pairs(OnCast.SpellAnnounce) do
                for evn, ev in pairs(sp) do
                    if evn ~= "meta" then
                        for chn, ch in pairs(ev.channels) do
                            if chn == "INSTANCE_CHAT" then
                                ev.channels["BATTLEGROUND"] = ch
                                ev.channels["INSTANCE_CHAT"] = nil
                                OnCast.printInfoMessage(sformat("%s %s channel 'INSTANCE_CHAT' changed to 'BATTLEGROUND'", spn, evn))
                            end
                        end
                    end
                end
            end
        end

        self.db.global.version = OnCast.VERSION -- Set updated version
    end
end

function OnCast:OnEnable()
   -- Set /commands
   self:RegisterChatCommand(COMMAND, "Command")
   self:RegisterChatCommand(MACRO_COMMAND, "MacroCommand")
   self:RegisterChatCommand(MACRO_COMMAND_SHORT, "MacroCommand")
   self:RegisterChatCommand(SPELL_COMMAND, "SpellCommand")
   self:RegisterChatCommand(DEBUG_COMMAND, "DebugCommand")
end

function OnCast:OnDisable()
    -- Unset /commands
    self:UnregisterChatCommand(COMMAND)
    self:UnregisterChatCommand(MACRO_COMMAND)
    self:UnregisterChatCommand(MACRO_COMMAND_SHORT)
    self:UnregisterChatCommand(SPELL_COMMAND)
    self:UnregisterChatCommand(DEBUG_COMMAND)
end

local CONDITION_PATTERN = "^%[(.+)%]$"

-- /oncast and /oncast-m
local function executeCommand(input, isMacro)
    -- /oncast [version][help [topic]] [@target,conditions] event[,event] channel[,exchannel(var[;var])] "Message" ...
    if not isMacro then
        local sysCmd, sysSub = OnCast:GetArgs(input, 2)
        if Console.parseSystemCommand(sysCmd, sysSub) then
            return
        end
    end

    -- Check for conditionals, e.g. [@target,btn:x,mod:shift,mod:ctrl,mod:alt]
    local conditionsArg, next = OnCast:GetArgs(input, 1)
    local conditionStr = conditionsArg and smatch(conditionsArg, CONDITION_PATTERN)
    local target, dynamicUnits = nil, nil
    if conditionStr then
        local ok, unitId = false, nil
        ok, unitId, dynamicUnits = MacroConditionals.Evaluate(conditionStr)
        if not ok then return end
        
        if unitId and WOW.UnitExists(unitId) then
            target = {id = unitId, name = WOW.GetUnitName(unitId,true), raidTargetIndex = WOW.GetRaidTargetIndex(unitId) }
        end
    else
        next = 1 -- Reset to start of input
    end

    -- Set announcement for the next spell cast
    local announce = OnCast.Announce or { meta = {} }
    local cmd, argEvents, argChannels, argMessage = {}, nil, nil, nil
    repeat
        -- Reset cmd
        cmd.meta = {}
        cmd.events = nil
        cmd.eventVars = nil
        cmd.channels = nil
        cmd.channelVars = nil
        cmd.conditions = nil
        cmd.message = nil

        -- Parse command set (events channels message)
        argEvents, argChannels, argMessage, next = OnCast:GetArgs(input, 3, next)
        if not Console.parseCommand(cmd, argEvents, argChannels, argMessage) then
            return
        end

        for i=1, #cmd.events do
            -- Create Event
            local eventName = cmd.events[i]
            local event = announce[eventName] or {}
            event.vars = cmd.eventVars[i]
            -- Create Channels
            if dynamicUnits then -- Conditional
                local conditionalChannels = event.conditionalChannels or {}
                local channels = {}
                for k=1, #cmd.channels do
                    local channelName = cmd.channels[k]
                    local channelVars = cmd.channelVars[k]
                    channels[channelName] = { message = cmd.message, vars = channelVars }
                end
                conditionalChannels[#conditionalChannels+1] = { channels = channels, units = dynamicUnits }
                event.conditionalChannels = conditionalChannels
            else -- Non-conditional
                local channels = event.channels or {}
                for k=1, #cmd.channels do
                    local channelName = cmd.channels[k]
                    local channelVars = cmd.channelVars[k]
                    channels[channelName] = {message = cmd.message, vars = channelVars}
                end
                event.channels = channels
            end
            -- Create Meta
            local eventMeta = event.meta or {skipStandard = true, skipExtended = true}
            eventMeta.skipStandard = eventMeta.skipStandard and cmd.meta.skipStandard
            eventMeta.skipExtended = eventMeta.skipExtended and cmd.meta.skipExtended
            eventMeta.target = target or eventMeta.target
            event.meta = eventMeta
            -- Set Event
            announce[eventName] = event
        end
        -- Create Announcement Meta
        if cmd.meta.hasExtendedEvents then
            announce.meta.hasExtendedEvents = true
        end
    until (next >= 1e9)

    -- Set timeout
    announce.timestamp = WOW.GetTime()
    announce.timeout = isMacro and 1 or 30 -- 1 second should be enough between button press and SENT
    -- Set Announcement
    OnCast.Announce = announce
end

function OnCast:Command(input)
    local status, err = pcall(executeCommand, input, false)
    if (not status) then
        print("Error while executing OnCast: \n" .. err)
    end
end

function OnCast:MacroCommand(input)
    local status, err = pcall(executeCommand, input, true)
    if (not status) then
        print("Error while executing OnCast-M: \n" .. err)
    end
end

-- Print channel messages for the registered spell and eventName
local function printSpellEvent(spell, eventName)
    if eventName == "meta" then return end
    -- Print event header
    local event = OnCast.SpellAnnounce[spell][eventName]
    if event.vars then
        local exEvent = OnCast.EXTENDED_EVENTS[eventName]
        local varArray = exEvent.VarsToArray(Util.tableDeepCopy(event.vars))
        local vars = tconcat(varArray, ";")
        OnCast.printInfoMessage(sformat("- %s(%s):", eventName, vars))
    else
        OnCast.printInfoMessage(sformat("- %s:", eventName))
    end
    -- Print channels
    for channelName,channel in pairs(event.channels) do
        if channel.vars then
            local exChannel = OnCast.EXTENDED_CHANNELS[channelName]
            local varArray = exChannel.VarsToArray(Util.tableDeepCopy(channel.vars))
            local vars = tconcat(varArray, ";")
            OnCast.printInfoMessage(sformat("-- %s(%s): \"%s\"", channelName, vars, channel.message))
        else
            OnCast.printInfoMessage(sformat("-- %s: \"%s\"", channelName, channel.message))
        end
    end
end

-- Shows the registed events for the given spell and eventNames, or all if nil
local function showRegisteredSpellEvents(spell, eventNames)
    -- Specific spell
    if spell then
        if OnCast.SpellAnnounce[spell] then
            if tonumber(spell) then
                OnCast.printInfoMessage(sformat("On %s (%s):", spell, WOW.GetSpellInfo(spell)), OnCast.HEADING_COLOR)
            else
                OnCast.printInfoMessage(sformat("On %s:", spell), OnCast.HEADING_COLOR)
            end
            -- Specific event
            if eventNames then 
                for i=1, #eventNames do
                    local eventName = eventNames[i]
                    if OnCast.SpellAnnounce[spell][eventName] then
                        printSpellEvent(spell, eventName)
                    else
                        OnCast.printInfoMessage(sformat(" - %s not registered", eventName))
                    end
                end
                return
            end
             -- All events
            for eventName in pairs(OnCast.SpellAnnounce[spell]) do
                printSpellEvent(spell, eventName)
            end
            return
        end
        OnCast.printInfoMessage(sformat("%s: No events registered", spell))
        return
    end
    -- All registered spells
    local hasSpells = false
    for spell in pairs(OnCast.SpellAnnounce) do
        if tonumber(spell) then
            OnCast.printInfoMessage(sformat("On %s (%s):", spell, WOW.GetSpellInfo(spell)), OnCast.HEADING_COLOR)
        else
            OnCast.printInfoMessage(sformat("On %s:", spell), OnCast.HEADING_COLOR)
        end
        for eventName in pairs(OnCast.SpellAnnounce[spell]) do 
            printSpellEvent(spell, eventName)
        end
        hasSpells = true
    end
    if not hasSpells then
        OnCast.printInfoMessage("No spell events registered", OnCast.HEADING_COLOR)
    end
end

local function removeRegisteredSpellEvents(spell, eventNames, channelNames)
    --print("Spell: "..spell)
    if not OnCast.SpellAnnounce[spell] then
        OnCast.printInfoMessage(sformat("No registered events found for %s", spell))
        return
    end
    
    if not eventNames then
        -- Remove all events
        OnCast.SpellAnnounce[spell] = nil
        OnCast.printInfoMessage(sformat("All events removed from %s", spell))
        return
    end

    for i=1, #eventNames do
        local eventName = eventNames[i]
        --print("Event: "..eventName)
        local event = OnCast.SpellAnnounce[spell][eventName]
        if event and eventName ~= "meta" then
            if channelNames then
                for k=1, #channelNames do
                    local channelName = channelNames[k]
                    --print("Channel: "..channel)
                    if event.channels and event.channels[channelName] then
                        event.channels[channelName] = nil
                        OnCast.printInfoMessage(sformat("%s removed from %s for %s", channelName, eventName, spell))
                    end
                end
                -- Check if event is now empty
                if Util.getTableSize(event.channels) == 0 then
                    OnCast.SpellAnnounce[spell][eventName] = nil
                    OnCast.printInfoMessage(sformat("-- Event %s removed from %s --", eventName, spell))
                    -- Check if spell is now empty
                    if Util.getTableSize(OnCast.SpellAnnounce[spell], {["meta"]=true}) == 0 then
                        OnCast.SpellAnnounce[spell] = nil
                        OnCast.printInfoMessage(sformat("-- All events removed from %s --", spell))
                    end
                else -- Revalidate event meta
                    local skipStandard, skipExtended = true, true
                    for channelName,_ in pairs(event.channels) do
                        if OnCast.STANDARD_CHANNELS[channelName] then
                            skipStandard = false
                        elseif OnCast.EXTENDED_CHANNELS[channelName] then
                            skipExtended = false
                        end
                    end
                    if not event.meta then event.meta = {} end
                    event.meta.skipStandard = skipStandard
                    event.meta.skipExtended = skipExtended
                end
            else -- Remove all channels
                OnCast.SpellAnnounce[spell][eventName] = nil
                OnCast.printInfoMessage(sformat("Event %s removed from %s", eventName, spell))
                -- Check if spell is now empty
                if Util.getTableSize(OnCast.SpellAnnounce[spell], {["meta"]=true}) == 0 then
                    OnCast.SpellAnnounce[spell] = nil
                    OnCast.printInfoMessage(sformat("-- All events removed from %s --", spell))
                else -- Revalidate spell meta
                    local hasExtendedEvents = false
                    for eventName,_ in pairs(OnCast.SpellAnnounce[spell]) do
                        if OnCast.EXTENDED_EVENTS[eventName] then
                            hasExtendedEvents = true
                        end
                    end
                    OnCast.SpellAnnounce[spell].meta.hasExtendedEvents = hasExtendedEvents and true or nil
                end
            end
        end
    end
end


local function executeSpellCommand(input)
    local max_args = 5 -- /oncast-spell [remove|show] "Spell" events channels "Message"
    local args = pack(OnCast:GetArgs(input, max_args))

    if Console.parseSpellSystemCommand(args) then
        return
    end
    
    local cmd = {
        meta = {},
        spell = nil,
        events = nil,
        channels = nil,
        channelVars = nil,
        message = nil,
        -- Flags
        remove = false,
        show = false,
    }

    if not Console.parseSpellCommand(cmd, args) then 
        return 
    end

    -- Show events for spell
    if cmd.show then
        showRegisteredSpellEvents(cmd.spell, cmd.events)
        return
    end

    -- Remove spell events
    if cmd.remove then   
        removeRegisteredSpellEvents(cmd.spell, cmd.events, cmd.channels)
        return
    end
        
    -- Set spell events
    local announce = OnCast.SpellAnnounce[cmd.spell] or {}
    for i=1, #cmd.events do
        -- Create Event
        local eventName = cmd.events[i]
        local event = announce[eventName] or {}
        event.vars = cmd.eventVars[i]
        -- Create Channels
        local channels = event.channels or {}
        for k=1, #cmd.channels do
            local channelName = cmd.channels[k]
            local channelVars = cmd.channelVars[k]
            channels[channelName] = {message = cmd.message, vars = channelVars}
        end
        event.channels = channels
        -- Create Event Meta
        local eventMeta = event.meta or {skipStandard = true, skipExtended = true}
        eventMeta.skipStandard = eventMeta.skipStandard and cmd.meta.skipStandard
        eventMeta.skipExtended = eventMeta.skipExtended and cmd.meta.skipExtended
        event.meta = eventMeta
        -- Set Event
        announce[eventName] = event
    end
    -- Create Announcement Meta
    local meta = announce.meta or {}
    if cmd.meta.hasExtendedEvents then
        meta.hasExtendedEvents = true
    end
    announce.meta = meta
    -- Set Announcement
    OnCast.SpellAnnounce[cmd.spell] = announce
    
    showRegisteredSpellEvents(cmd.spell)
end

function OnCast:SpellCommand(input)
    local status, err = pcall(executeSpellCommand, input)
    if (not status) then
        print("Error while executing OnCast-Spell: \n" .. err)
    end
end

local function executeDebug(input)
    local max_args = 5 -- /oncast-debug [on/off]
    local args = pack(OnCast:GetArgs(input, max_args))

    local arg1 = #args>1 and args[1]

    if arg1 then
        if arg1 == "on" then
            OnCast.debug = true
        elseif arg1 == "off" then
            OnCast.debug = false
        elseif arg1 == "memory" then
            local inUse = collectgarbage("count")
            inUse = inUse / 1000 -- convert to kB
            OnCast.printInfoMessage(sformat("Memory Use: %0.2fkB", inUse))
            OnCast:GetModule("Event").DebugLists()
            OnCast:GetModule("ExtEvent").DebugLists()
            return
        end
    else    
        OnCast.debug = (not OnCast.debug)
    end

    OnCast.printDebugMessage(OnCast.debug and "ON" or "OFF")
end

function OnCast:DebugCommand(input)
    local status, err = pcall(executeDebug, input)
    if (not status) then
        print("Error while executing OnCast-Debug: \n" .. err)
    end
end
